#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

import logging
import sys


def setup_logging(debug=False):
    # This all is a bit of hack and should be replaced with a proper logging configuration

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG if debug else logging.WARNING)

    # Clear existing handlers to prevent duplicate logs
    while logger.handlers:
        logger.handlers.pop()

    # Define a common formatter
    formatter = logging.Formatter(
        "%(name)s: %(asctime)s | %(levelname)s | %(filename)s:%(lineno)s | %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )

    # Add StreamHandler for console output in debug mode
    if debug:
        debug_handler = logging.StreamHandler(sys.stdout)
        debug_handler.setLevel(logging.DEBUG)
        debug_handler.setFormatter(formatter)
        logger.addHandler(debug_handler)

    # Always add a StreamHandler for warnings and above
    warning_handler = logging.StreamHandler(sys.stdout)
    warning_handler.setLevel(logging.WARNING)
    warning_handler.setFormatter(formatter)
    logger.addHandler(warning_handler)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

from .cli import cli

cli()



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

from collections import deque
from typing import Any, List, Tuple, Union

import grpc
import numpy as np
from dataclasses import dataclass

from featureform import Model
from featureform.proto import serving_pb2


@dataclass
class TrainTestSplitDetails:
    name: str
    version: str
    model: Union[str, Model]
    test_size: float
    shuffle: bool
    random_state: int
    batch_size: int = 1

    def to_proto(
        self, request_type: serving_pb2.RequestType
    ) -> serving_pb2.TrainTestSplitRequest:
        req = serving_pb2.TrainTestSplitRequest()
        req.id.name = self.name
        req.id.version = self.version
        if self.model is not None:
            req.model.name = (
                self.model if isinstance(self.model, str) else self.model.name
            )
        req.test_size = self.test_size
        req.shuffle = self.shuffle
        req.random_state = self.random_state
        req.request_type = request_type
        req.batch_size = self.batch_size
        return req


class _SplitStream:
    """
    Handles the gRPC stream for the train test split
    """

    def __init__(self, stub, train_set_split_details: TrainTestSplitDetails):
        self.request_queue = deque()
        self.response_stream = None
        self.stub = stub
        self.train_test_split_details = train_set_split_details

    def start(self):
        """Initializes the gRPC stream."""
        self.response_stream = self.stub.TrainTestSplit(
            self._create_request_generator()
        )
        init_response = self.send_request(serving_pb2.RequestType.INITIALIZE)
        if not init_response.initialized:
            raise ValueError("Failed to initialize train test split")
        return self

    def send_request(self, request_type) -> serving_pb2.BatchTrainTestSplitResponse:
        """Sends a request through the stream and returns the response."""
        req = self.train_test_split_details.to_proto(request_type)
        self.request_queue.append(req)
        try:
            response = next(self.response_stream)
            return response
        except grpc.RpcError as e:
            print(f"Caught RPC error of type: {e.code()} - {e.details()}")
            if e.code() == grpc.StatusCode.RESOURCE_EXHAUSTED:
                print("Message size exceeded. Try reducing the batch size.")
            raise

    def _create_request_generator(self):
        """Yields requests from the queue for the gRPC stream."""
        while True:
            if self.request_queue:
                yield self.request_queue.popleft()


@dataclass
class _NpProtoTypeParser:
    """Parses the proto types to numpy types"""

    nptype: np.dtype
    feature_proto_types: List[Tuple[str, np.dtype]]
    label_proto_type: str

    @staticmethod
    def init_types(row: serving_pb2.TrainingDataRow):
        from featureform.serving import (
            get_numpy_array_type,
            proto_type_to_np_type,
        )

        types = [
            (feature.WhichOneof("value"), proto_type_to_np_type(feature))
            for feature in row.features
        ]
        numpy_array_type = get_numpy_array_type(list(map(lambda x: x[1], types)))
        label_proto_type = row.label.WhichOneof("value")
        return _NpProtoTypeParser(numpy_array_type, types, label_proto_type)

    def parse_features(self, rows: List[serving_pb2.Value]) -> List[Any]:
        return [
            getattr(feature, self.feature_proto_types[i][0])
            for i, feature in enumerate(rows)
        ]

    def parse_label(self, row: serving_pb2.Value) -> Any:
        return getattr(row, self.label_proto_type)


class TrainTestSplitIterator:
    def __init__(
        self,
        split_stream: _SplitStream,
        batch_size: int,
        request_type: serving_pb2.RequestType,
    ):
        self.request_type = request_type
        self._split_stream = split_stream
        self._batch_size = batch_size
        self._complete = False
        self._np_type_parser = None

    @staticmethod
    def train_iter(split_stream, batch_size):
        return TrainTestSplitIterator(
            split_stream, batch_size, serving_pb2.RequestType.TRAINING
        )

    @staticmethod
    def test_iter(split_stream, batch_size):
        return TrainTestSplitIterator(
            split_stream, batch_size, serving_pb2.RequestType.TEST
        )

    def __iter__(self):
        return self

    def __next__(self) -> Tuple[np.ndarray, np.ndarray]:
        if self._complete:
            raise StopIteration

        batch_features_list = []
        batch_labels_list = []

        result = self._split_stream.send_request(self.request_type)
        data = result.data

        if self._np_type_parser is None and data.rows:
            self._np_type_parser = _NpProtoTypeParser.init_types(data.rows[0])

        for row in data.rows:
            batch_features_list.append(
                self._np_type_parser.parse_features(row.features)
            )
            batch_labels_list.append(self._np_type_parser.parse_label(row.label))

        if result.iterator_done:
            self._complete = True

        if not batch_features_list:  # If no data was collected
            raise StopIteration

        batch_features = np.array(
            batch_features_list, dtype=self._np_type_parser.nptype
        )
        batch_labels = np.array(batch_labels_list)

        return batch_features, batch_labels


class TrainTestSplit:
    """
    Returns two iterators, one for the training set and one for the test set, in that order
    """

    def __init__(
        self,
        stub,
        name: str,
        version: str,
        test_size: float,
        train_size: float,  # unused for now
        shuffle: bool,
        random_state: int,
        batch_size: int,
        model: Union[str, Model] = None,
    ):
        self.batch_size = batch_size
        self.train_iter = None
        self.test_iter = None

        self.train_test_split_details = TrainTestSplitDetails(
            name=name,
            version=version,
            model=model,
            test_size=test_size,
            shuffle=shuffle,
            random_state=random_state,
            batch_size=batch_size,
        )

        self._split_stream = _SplitStream(stub, self.train_test_split_details).start()

    def split(self):
        return (
            TrainTestSplitIterator.train_iter(self._split_stream, self.batch_size),
            TrainTestSplitIterator.test_iter(self._split_stream, self.batch_size),
        )



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

import pytest

from featureform import FeatureColumnResource


# conftest.py
# 1. data_dictionary() (fixture) -> {columns: [""], data: [[],[]]}
# 2. features_dataframe(data_dictionary) (fixture) -> pandas dataframe from dictionary
# 3. primary_dataset() (fixture) -> ColumnSourceRegistrar() fall the register_file method and see the return type
# 4. MultiFeature(features_dataframe, features_dataset) (fixture) -> MultiFeature object with dataset and dataframe


def test_multi_feature(multi_feature):
    assert multi_feature.tags == []


@pytest.mark.parametrize(
    "input_str, expected_str",
    [
        ("column1", "column1"),
        ('"column2"', "column2"),
        ('"column3', "column3"),
        ('""', ""),
    ],
)
def test_clean_name(multi_feature, input_str, expected_str):
    result = multi_feature._clean_name(input_str)
    assert result == expected_str


@pytest.mark.parametrize(
    "input_str, has_quotes, expected_str",
    [
        ("column1", True, '"column1"'),
        ('"column2"', True, '"column2"'),
        ('""', True, '""'),
        ('"column3', False, "column3"),
        ('""', False, ""),
    ],
)
def test_modify_column_name(multi_feature, input_str, has_quotes, expected_str):
    result = multi_feature._modify_column_name(input_str, has_quotes)
    assert result == expected_str


@pytest.mark.parametrize(
    "include_cols, exclude_cols, expected_list",
    [
        ([], [], ["a", "b", "c", "ts"]),
        (["a"], ["ts"], ["a"]),
        (["a", "b", "ts"], ["c"], ["a", "b", "ts"]),
        ([], ["b", "c"], ["a", "ts"]),
    ],
)
def test_get_feature_columns_no_ts(
    include_cols, exclude_cols, expected_list, multi_feature, features_dataframe
):
    result = multi_feature._get_feature_columns(
        df=features_dataframe,
        include_columns=include_cols,
        exclude_columns=exclude_cols,
        entity_column="entity",
        timestamp_column="",
    )
    assert set(result) == set(expected_list)


@pytest.mark.parametrize(
    "include_cols, exclude_cols, expected_list",
    [
        ([], [], ["a", "b", "c"]),
        (["a"], [], ["a"]),
        (["a", "b"], ["c"], ["a", "b"]),
        ([], ["b", "c"], ["a"]),
    ],
)
def test_get_feature_columns(
    include_cols, exclude_cols, expected_list, multi_feature, features_dataframe
):
    result = multi_feature._get_feature_columns(
        df=features_dataframe,
        include_columns=include_cols,
        exclude_columns=exclude_cols,
        entity_column="entity",
        timestamp_column="ts",
    )
    assert set(result) == set(expected_list)


@pytest.mark.parametrize(
    "include_cols, exclude_cols",
    [
        (["a", "b", "c"], ["c"]),
        (["a", "d"], []),
        ([], ["c", "d"]),
    ],
)
def test_get_feature_columns_errors(
    include_cols, exclude_cols, multi_feature, features_dataframe
):
    with pytest.raises(ValueError):
        multi_feature._get_feature_columns(
            df=features_dataframe,
            include_columns=include_cols,
            exclude_columns=exclude_cols,
            entity_column="entity",
            timestamp_column="ts",
        )


def test_multi_feature_is_iterable(multi_feature):
    for feature in multi_feature:
        assert isinstance(feature, FeatureColumnResource)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

import os


def is_enabled(feature_name, default=False):
    """
    Returns True if the feature is enabled, currently using an environment variable.
    Interprets the environment variable as a boolean if it's a valid representation, else defaults to the specified default.
    """
    value = os.getenv(feature_name)
    if value is not None:
        return value.lower() == "true"
    return default



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

import pytest
from featureform.register import ModelRegistrar
from featureform.resources import Model
import featureform as ff


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_getting_model_successfully(is_local, is_insecure):
    model_name = "model_i"

    resource_client = arrange_resources(model_name, is_local, is_insecure)

    model = resource_client.get_model(model_name, is_local)

    assert (
        isinstance(model, Model)
        and model.name == model_name
        and model.get_resource_type() == ff.ResourceType.MODEL
    )


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_getting_model_by_unregistered_name(is_local, is_insecure):
    model_name = "model_j"

    resource_client = arrange_resources(model_name, is_local, is_insecure)
    if is_local:
        with pytest.raises(IndexError, match="out of range"):
            resource_client.get_model("model_z", is_local)
    else:
        model = resource_client.get_model("model_z", is_local)
        assert model is None


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_getting_model_no_name(is_local, is_insecure):
    model_name = "model_k"

    resource_client = arrange_resources(model_name, is_local, is_insecure)

    with pytest.raises(
        TypeError, match="missing 1 required positional argument: 'name'"
    ):
        resource_client.get_model(local=is_local)


@pytest.fixture(autouse=True)
def before_and_after_each(setup_teardown):
    setup_teardown()
    yield
    setup_teardown()


def arrange_resources(model_name, is_local, is_insecure):
    ff.register_model(model_name)
    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply(asynchronous=True)

    return resource_client



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

# For fetching sources as dataframes, the default limit is set to -1
# to denote that all records should be fetched.
NO_RECORD_LIMIT = -1

# 2 million records
TWO_MILLION_RECORD_LIMIT = 2_000_000



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#  Copyright 2024 FeatureForm Inc.
#

import time
import featureform as ff
import pytest


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_registering_resources_out_of_order(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    training_set_name = "fraud_training"
    training_set_variant = "quickstart"

    # Arranges the resources context following the Quickstart pattern
    arrange_resources_out_of_order(
        provider,
        source,
        inference_store,
        is_local,
        training_set_name,
        training_set_variant,
    )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply(asynchronous=True)
    wait_for_training_set(
        resource_client, training_set_name, training_set_variant, is_local
    )

    training_set = resource_client.print_training_set(
        training_set_name, training_set_variant, is_local
    )

    if is_local:
        assert training_set is not None and training_set["status"] == "ready"
    else:
        assert (
            training_set is not None
            and training_set.status.Status._enum_type.values[
                training_set.status.status
            ].name
            == "READY"
        )


@pytest.fixture(autouse=True)
def before_and_after_each(setup_teardown):
    setup_teardown()
    yield
    setup_teardown()


def arrange_resources_out_of_order(
    provider, source, online_store, is_local, training_set_name, training_set_variant
):
    ff.register_training_set(
        training_set_name,
        training_set_variant,
        label=("fraudulent", "quickstart"),
        features=[("avg_transactions", "quickstart")],
    )

    if is_local:

        @provider.df_transformation(
            variant="quickstart", inputs=[("transactions", "quickstart")]
        )
        def average_user_transaction(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(variant="quickstart")
        def average_user_transaction():
            return "SELECT customerid as user_id, avg(transactionamount) as avg_transaction_amt from {{transactions.quickstart}} GROUP BY user_id"

    feature_column = "TransactionAmount" if is_local else "avg_transaction_amt"
    label_column = "IsFraud" if is_local else "isfraud"
    inference_store = provider if is_local else online_store

    average_user_transaction.register_resources(
        entity="user",
        entity_column="CustomerID" if is_local else "user_id",
        inference_store=inference_store,
        features=[
            {
                "name": "avg_transactions",
                "variant": "quickstart",
                "column": feature_column,
                "type": "float32",
            },
        ],
    )

    # register feature with no inference store
    average_user_transaction.register_resources(
        entity="user",
        entity_column="CustomerID" if is_local else "user_id",
        features=[
            {
                "name": "avg_transactions2",
                "variant": "quickstart",
                "column": feature_column,
                "type": "float32",
            },
        ],
    )

    source.register_resources(
        entity="user",
        entity_column="CustomerID" if is_local else "customerid",
        labels=[
            {
                "name": "fraudulent",
                "variant": "quickstart",
                "column": label_column,
                "type": "bool",
            },
        ],
    )

    ff.register_entity("user")


def wait_for_training_set(
    resource_client, training_set_name, training_set_variant, is_local
):
    if not is_local:
        start = time.time()
        while True:
            time.sleep(3)
            ts = resource_client.get_training_set(
                training_set_name, training_set_variant
            )
            elapsed_wait = time.time() - start
            if (elapsed_wait >= 60) and ts.status != "READY":
                print(
                    f"Wait time for training set status exceeded; status is {ts.status}"
                )
                break
            elif ts.status == "READY":
                print(f"Training set is ready")
                break
            else:
                print(
                    f"Training set status is currently {ts.status} after {elapsed_wait} seconds ..."
                )
                continue
